<?php  // $Id$

// dbpositions enrollment plugin is based on the contrib/dbuserrel
// plugin, with the addition of syncing position assignments

require_once($CFG->dirroot.'/enrol/enrol.class.php');

class enrolment_plugin_dbpositions {

    var $log;

/*
 * For the given user, let's go out and look in an external database
 * for an authoritative list of relationships, and then adjust the
 * local Moodle assignments to match.
 */
function setup_enrolments(&$user=null) {
    global $CFG;

    // Store the field values in some shorter variable names to ease reading of the code.
    $flocalsubject  = $CFG->enrol_dbpositions_localsubjectuserfield;
    $flocalobject   = $CFG->enrol_dbpositions_localobjectuserfield;
    $fremotesubject = $CFG->enrol_dbpositions_remotesubjectuserfield;
    $fremoteobject  = $CFG->enrol_dbpositions_remoteobjectuserfield;
    $dbtable        = $CFG->enrol_dbpositions_table;
    $fremotepos = $CFG->enrol_dbpositions_remoteposfield;
    $flocalpos = $CFG->enrol_dbpositions_localposfield;
    $fremoteorg = $CFG->enrol_dbpositions_remoteorgfield;
    $flocalorg = $CFG->enrol_dbpositions_localorgfield;
    $fpostype = $CFG->enrol_dbpositions_postypefield;
    $ffullname = $CFG->enrol_dbpositions_fullnamefield;
    $fshortname = $CFG->enrol_dbpositions_shortnamefield;

    //The shortname of the role that we will apply to <SubjectUser>
    //in the user-level context of <ObjectUser>:
    $roleshortname = $CFG->enrol_dbpositions_roleshortname;

    // NOTE: if $this->enrol_connect() succeeds you MUST remember to call
    // $this->enrol_disconnect() as it is doing some nasty vodoo with $CFG->prefix
    $enroldb = $this->enrol_connect();
    if (!$enroldb) {
        error_log('Error: [ENROL_DBPOSITIONS] Could not make a connection');
        return;
    }

    $role = get_record('role', 'shortname', $roleshortname);
    if ($role) {
        $roleid = $role->id;

        if ($user) {
            $subjectfield = $enroldb->quote($user->{$flocalsubject});
            $objectfield = $enroldb->quote($user->{$flocalobject});

            $sql = "SELECT * FROM {$dbtable}
                WHERE {$fremotesubject} = $subjectfield
                OR {$fremoteobject} = $objectfield";
        } else {
            $sql = "SELECT * FROM {$dbtable}";
        }

        if ($rs = $enroldb->Execute($sql)) {
            $sql = "SELECT " . sql_concat_join("'|'", array("u1.{$flocalsubject}", "u2.{$flocalobject}")) . " AS unique,
                ra.*,
                u1.{$flocalsubject} AS subjectid,
                u2.{$flocalobject} AS objectid
                FROM {$CFG->prefix}role_assignments ra
                JOIN {$CFG->prefix}context c ON c.id = ra.contextid
                JOIN {$CFG->prefix}user u1 ON ra.userid = u1.id
                JOIN {$CFG->prefix}user u2 ON c.instanceid = u2.id
                WHERE c.contextlevel = " . CONTEXT_USER .
                " AND ra.enrol='dbpositions'" .
                (!empty($user) ?  " AND c.instanceid = {$user->id} OR ra.userid = {$user->id}" : '');
            echo $sql;

            if (!$existing = get_records_sql($sql)) {
                $existing = array();
            }

            if (!$rs->EOF) {

                $subjectusers = array(); // cache of mapping of localsubjectuserfield to mdl_user.id (for get_context_instance)
                $objectusers = array(); // cache of mapping of localobjectuserfield to mdl_user.id (for get_context_instance)
                $contexts = array(); // cache

                $rels = array();
                while ($row = rs_fetch_next_record($rs)) {
                    // either we're assigning ON the current user, or TO the current user
                    $key = $row->{$fremotesubject} . '|' . $row->{$fremoteobject};
                    if (array_key_exists($key, $existing)) {
                        // exists in moodle db already, unset that row in the moodle array
                        unset($existing[$key]);
                        continue;
                    }

                    if (!array_key_exists($row->{$fremotesubject}, $subjectusers)) {
                        $subjectusers[$row->{$fremotesubject}] = get_field('user', 'id', $flocalsubject, $row->{$fremotesubject});
                    }
                    if ($subjectusers[$row->{$fremotesubject}] == false) {
                        error_log("Warning: [" . $row->{$fremotesubject} . "] couldn't find subject user -- skipping $key");
                        // couldn't find user, skip
                        continue;
                    }

                    if (!array_key_exists($row->{$fremoteobject}, $objectusers)) {
                        $objectusers[$row->{$fremoteobject}] = get_field('user', 'id', $flocalobject, $row->{$fremoteobject});
                    }
                    if ($objectusers[$row->{$fremoteobject}] == false) {
                        // couldn't find user, skip
                        error_log("Warning: [" . $row->{$fremoteobject} . "] couldn't find object user --  skipping $key");
                        continue;
                    }
                    $context = get_context_instance(CONTEXT_USER, $objectusers[$row->{$fremoteobject}]);
                    error_log("Information: [" . $row->{$fremotesubject} . "] assigning $roleshortname" .
                            " to user with $flocalsubject" . $row->{$fremotesubject} .
                            " on context of user with $fremoteobject " . $row->{$fremoteobject});
                    role_assign($roleid, $subjectusers[$row->{$fremotesubject}], 0, $context->id, 0, 0, 0, 'dbpositions');
                }

                // delete everything left in existing array
                foreach ($existing as $key => $assignment) {
                    if ($assignment->enrol == 'dbpositions') {
                        error_log("Information: [$key] unassigning $key");
                        role_unassign($assignment->roleid, $assignment->userid, 0, $assignment->contextid);
                    }
                }
            } else {
                error_log('Warning: [ENROL_DBPOSITIONS] Couldn\'t get rows from external db: '.$enroldb->ErrorMsg(). ' -- no relationships to assign');
            }
        }
    }

    // If site admin has not speicifed a field in the external database table which identifies which
    // type of position the row relates to, assume that we're only dealing in primary positions
    if (empty($fpostype)) {
        $typesarray = array('primary' => 1);
    } else {
        $typesarray = array('primary' => 1, 'secondary' => 2, 'aspirational' => 3);
    }

    //Prepare some data needed to insert positions
    $orgsql = 'SELECT ' . addslashes($flocalorg) . ',id FROM '. $CFG->prefix . 'org';
    $organisations = get_records_sql($orgsql);
    $possql = 'SELECT ' . addslashes($flocalpos) . ',id FROM '. $CFG->prefix . 'pos';
    $positions = get_records_sql($possql);
    $userssql = 'SELECT ' . addslashes($flocalobject) . ',id FROM ' . $CFG->prefix . 'user';
    $users = get_records_sql($userssql);
    $reportingsql = 'SELECT ' . sql_concat_join("'|'", array('u1.' .
        addslashes($flocalsubject), "u2." . addslashes($flocalobject))) . ', ra.id ' .
            'FROM ' . $CFG->prefix . 'role_assignments ra ' .
            ' INNER JOIN ' . $CFG->prefix . 'user u1 on u1.id=ra.userid' .
            ' INNER JOIN ' . $CFG->prefix . 'context c on c.id = ra.contextid' .
            ' INNER JOIN ' . $CFG->prefix . 'user u2 on c.instanceid = u2.id ' .
            'WHERE c.contextlevel = ' . CONTEXT_USER;
    echo $reportingsql;
    $reportinglines = get_records_sql($reportingsql);
    foreach ($typesarray as $positiontype => $positionnum) {
        $moodlepasql = 'SELECT u.' . addslashes($flocalobject) . ' as junkkey, pa.* ' .
                'FROM ' . $CFG->prefix . 'pos_assignment pa' .
                ' INNER JOIN ' . $CFG->prefix . 'user u on u.id=pa.userid';
        $externalpasql = 'SELECT * FROM ' . addslashes($dbtable);
        $externalpawhere = array();
        if ($user) {
            $externalpawhere[] = "$fremoteobject = '" . $user->$flocalobject. "'";
        }

        #Limit our queries to only pick up position assignments of the right type.
        #If no remote positiontype field has been specified, we assume that all external rows
        #relate to a primary position assignment
        $moodlepasql .= " WHERE pa.type = $positionnum";
        if ($user) {
            $moodlepasql .= " and u.id = " . $user->id;
        }
        if ($fpostype) {
            $externalpawhere[] = "$fpostype = $positiontype";
        }
        if ($externalpawhere) {
            $externalpasql .= ' WHERE (' . implode(') AND (', $externalpawhere) . ')';
        }
        # Retreive position assignments according to the external db:
        if ($externalpas = $enroldb->Execute($externalpasql)) {
            $positionstats['inserted'] = 0;
            $positionstats['updated'] = 0;
            $positionstats['deleted'] = 0;
            $positionstats['unchanged'] = 0;
            $positionstats['errors'] = 0;
            # Obtain existing moodle position assignments:
            $moodlepas = get_records_sql($moodlepasql);
            if (empty($moodlepas)) {
                $moodlepas = array();
            }

            while ($externalpa = rs_fetch_next_record($externalpas)) {
                $wellformedpa = true; // Optimism
                $positionassignment->fullname = $externalpa->{$ffullname};
                $positionassignment->shortname = $externalpa->{$fshortname};
                $positionassignment->organisationid = $organisations[$externalpa->{$fremoteorg}]->id;
                $positionassignment->userid = $users[$externalpa->{$fremoteobject}]->id;
                $positionassignment->positionid = $positions[$externalpa->{$fremotepos}]->id;
                $reportingkey = $externalpa->{$fremotesubject} . "|" . $externalpa->{$fremoteobject};
                $positionassignment->reportstoid = $reportinglines[$reportingkey]->id;
                $positionassignment->type = $positionnum;
                if (empty($positionassignment->organisationid) || empty($positionassignment->userid) || empty($positionassignment->positionid)){
                    $wellformedpa = false;
                }

                if (array_key_exists($externalpa->{$fremoteobject},$moodlepas)) {
                    $pakeys = array('fullname','shortname','organisationid',
                                    'userid', 'positionid', 'reportstoid');
                    $updateneeded = false;
                    foreach ($pakeys as $pakey) {
                        if ($positionassignment->{$pakey} != $moodlepas[$externalpa->{$fremoteobject}]->{$pakey}) {
                            $updateneeded = true;
                            break;
                        }
                    }
                    $un = ($updateneeded) ? 'yes':'no';
                    error_log('UPDATED NEEDED?: '.$un);
                    if ($updateneeded) {
                        if ($wellformedpa) {
                            $positionassignment->id = $moodlepas[$externalpa->{$fremoteobject}]->id;
                            $positionassignment->timemodified = time();
                            $positionassignment->usermodified = 0;
                            error_log('Information: Updating pa for ' . $fremoteobject . ':' . $externalpa->{$fremoteobject});
                            update_record('pos_assignment',addslashes_object($positionassignment));
                            $positionstats['updated']++;
                        } else {
                            error_log('Error: Trying to update pa for ' . $fremoteobject . ':' . $externalpa->{$fremoteobject} .
                                    ' with poorly formed data - skipping');
                            $positionstats['errors']++;
                        }
                    } else {
                        $positionstats['unchanged']++;
                    }
                    unset($moodlepas[$externalpa->{$fremoteobject}]);
                } else {
                    if ($wellformedpa) {
                        $positionassignment->timemodified = time();
                        $positionassignment->timecreated = time();
                        $positionassignment->usermodified = 0;
                        error_log('Information: Inserting pa for ' . $fremoteobject . ':' . $externalpa->{$fremoteobject});
                        insert_record('pos_assignment', addslashes_object($positionassignment));
                        $positionstats['inserted']++;
                    } else {
                        error_log('Error: Trying to insert pa for ' . $fremoteobject . ' ' . $externalpa->{$fremoteobject} .
                                ' with poorly formed data - skipping');
                        $positionstats['errors']++;
                    }
                }
            }
            if (!empty($moodlepas)) {
                foreach ($moodlepas as $moodlepa) {
                    delete_records('pos_assignment','id',$moodlepa->id);
                    $positionstats['deleted']++;
                }
            }
        }
    }
    $this->enrol_disconnect($enroldb);
    if (!$user) {
        error_log("Position assignment summary:");
        foreach ($positionstats as $key => $value) {
            error_log("Position assignment $key: $value");
        }
    }
}


/// Overide the get_access_icons() function
function get_access_icons($course) {
}


/// Overide the base config_form() function
function config_form($frm) {
    global $CFG;

    $vars = array(
        'type', 'host', 'user', 'pass', 'name', 'useenroldatabase',
        'useauthdb', 'table', 'localsubjectuserfield', 'localobjectuserfield',
        'remotesubjectuserfield', 'remoteobjectuserfield', 'roleshortname',
        'remoteposfield','localposfield',
        'remoteorgfield','localorgfield',
        'postypefield','shortnamefield', 'fullnamefield',
        );

    foreach ($vars as $var) {
        $var = 'enrol_dbpositions_' . $var;
        if (!isset($frm->$var)) {
            $frm->$var = '';
        }
    }
    include("$CFG->dirroot/enrol/dbpositions/config.html");
}

/// Override the base process_config() function
function process_config($config) {
    foreach ((array)$config as $key => $value) {
        if (strpos($key, 'enrol_dbpositions_') !== 0) {
            continue;
        }
        set_config($key, $value);
    }

    return true;
}

function enrol_connect() {
    global $CFG;

    // make up settings!
    $type = '';
    $host = '';
    $user = '';
    $pass = '';
    $name = '';
    $copyobj = $CFG;
    $copyprefix = 'enrol_dbpositions_';
    if (!empty($CFG->enrol_dbpositions_userenroldatabase)) {
        $copyprefix = 'enrol_db';
    }
    else if (!empty($CFG->enrol_dbpositions_useauthdb)) {
        $copyobj = get_config('auth/db');
        $copyprefix = '';
    }
    foreach (array('type', 'host', 'user', 'pass', 'name') as $setting) {
        $$setting = $copyobj->{$copyprefix . $setting};
    }
    // Try to connect to the external database (forcing new connection)
    $enroldb = &ADONewConnection($type);
    if ($enroldb->Connect($host, $user, $pass, $name, true)) {
        $enroldb->SetFetchMode(ADODB_FETCH_ASSOC); ///Set Assoc mode always after DB connection
        return $enroldb;
    } else {
        trigger_error("Error connecting to enrolment DB backend with: "
                      . "$host,$user,$pass,$name");
        return false;
    }
}

/// DB Disconnect
function enrol_disconnect($enroldb) {
    global $CFG;

    $enroldb->Close();
}

} // end of class


?>
